from datetime import datetime
import logging
from os import remove
from pathlib import Path
from shutil import copy
from typing import Optional

from mats.test import Test


class ArchiveManager:
    """
    The default data manager that saves data in a text format which is very flexible, \
    yet maintains simplicity and compatibility with analysis tools such as pandas.

    :param path: a string or `Path` containing the path to the data file
    :param fname: data file name
    :param delimiter: the data delimiter
    :param data_format: an integer that determines which data format
    :param preamble: a string that may be appended to the beginning of a data file
    :param loglevel: the logging level
    """

    def __init__(
        self,
        path: str | Path = ".",
        fname: str = "data.txt",
        delimiter: str = "\t",
        data_format: int = 0,
        preamble: Optional[str] = None,
        loglevel=logging.INFO,
    ):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._logger.setLevel(loglevel)

        self._fname = fname
        self._path = Path(path)
        self._delimiter = delimiter
        self._format = data_format
        self._preamble = preamble

    def aggregate(
        self,
        datetime: datetime,
        is_passing: bool,
        failed: list[str],
        tests: list[Test],
    ):
        """
        Collects data set from provided information and from the tests \
        and organizes the data into a dict in preparation for saving \
        a row of data.

        :param datetime: the datetime that the test started
        :param is_passing: True/False/None indicating if the test passed, \
        failed, or was aborted
        :param failed: a list of test monikers that failed
        :param tests: a list of tests representing the test sequence
        :return:
        """
        test_data = {
            "datetime": {"value": datetime},
            "pass": {"value": is_passing},
            "failed": {"value": failed.copy()},
        }

        for t in tests:
            test_data[t.moniker] = {"value": t.value}
            if t.criteria is not None:
                test_data[t.moniker]["criteria"] = t.criteria
            for k, v in t.saved_data.items():
                test_data[k]["value"] = v

        self.save(test_data)

    def save(self, point: dict):
        """
        Data save function

        Takes a point of data supplied as a dict and, depending on existing
        conditions, will archive the data point on the disk.  Each ``point``
        represents a single row of data representing the execution of a
        single instance of the test execution.  Each key of the ``dict`` has
        will represent the parameter being measured.

        Three conditions possible at save time:

         * Data file does not exist
         * Data file exists and is compatible with the current data point \
         (the headings and header strings match)
         * Data file exists, but is not compatible with the current data \
         point (the heading and header strings do not match)

        When the data file does not exist, then the ``save()`` method will
        create the new data file at ``<fname>.<extension>``.

        When the data file exists at ``<fname>.<extension>`` and is compatible
        with the current data point, then the ``save()`` method will simply
        append the new data point to previous data in a tabular data_format.

        Finally, when the current data point is deemed incompatible with the
        previous data, then the ``save()`` method will copy the old file to
        ``<fname>_<date>.<extension>`` and then create a new file at
        ``<fname>.<extension>`` under which data will be collected until a new
        data_format is once again detected.

        :param point: a ``dict`` containing key: value pairs which specify \
        the data to be saved in {'heading': {'value': value}}; the inner \
        dictionary may also contain a ``dict`` called ``criteria`` which \
        will contain the ``pass_if``, ``min``, or ``max`` values allowed
        :return: None
        """
        if self._format == 0:
            self._save_fmt0(point)
        elif self._format == 1:
            self._save_fmt1(point)
        else:
            raise ValueError(f'data_format "{self._format}" invalid')

    def _save_fmt0(self, point: dict):
        """
        Saves data with pass/fail criteria embedded into each header according
        to "data_format 0"

        :param point: a dict containing the name, value, and pass/fail \
        criteria.
        """
        headers = list(point.keys())

        heading_string = ""
        for header in headers:
            heading_fragments = []
            criteria = point[header].get("criteria")
            if criteria is not None:
                if "pass_if" in criteria.keys():
                    heading_fragments.append(f"pass_if={criteria['pass_if']}")
                if "min" in criteria.keys():
                    heading_fragments.append(f"min={criteria['min']}")
                if "max" in criteria.keys():
                    heading_fragments.append(f"max={criteria['max']}")

                heading_string += header + ":" + ",".join(heading_fragments) + "\n"

        preamble = '' if not self._preamble else f'{self._preamble}\n'
        header_string = preamble \
                        + heading_string \
                        + f"\n{self._delimiter.join(headers)}\n"

        data = []
        for _, value in point.items():
            v = value.get("value")
            if isinstance(v, str):
                data.append(v)
            elif isinstance(v, int):
                data.append(f"{v}")
            elif isinstance(v, float):
                data.append(f"{v}")
            else:
                try:
                    # convert from pint-style values
                    data.append(f"{v.magnitude}")
                except AttributeError:
                    data.append(str(v))  # this is the catch-all

        data_string = f"{self._delimiter.join(data)}\n"

        self._save_file(header_string, data_string)

    def _save_fmt1(self, point: dict):
        """
        Saves data with pass/fail criteria embedded into each header according
        to "data_format 1"

        :param point: a dict containing the name, value, and pass/fail \
        criteria.
        """
        headers = list(point.keys())

        header_string = ""
        for header in headers:
            header_string += f"{header}\t"
            criteria = point[header].get("criteria")

            if criteria is not None:
                if "pass_if" in criteria.keys():
                    header_string += f"{header} =\t"
                if "min" in criteria.keys():
                    header_string += f"{header} >=\t"
                if "max" in criteria.keys():
                    header_string += f"{header} <=\t"

        preamble = '' if not self._preamble else f'{self._preamble}\n\n'
        header_string = preamble \
                        + header_string.strip() \
                        + "\n"

        data = []
        for header, value in point.items():
            v = value.get("value")
            if header == "failed":
                # when header is "failed", then save each failed value into
                # its own special semicolon-separated data_format
                data.append(";".join(v))
            else:
                if isinstance(v, str):
                    data.append(v)
                elif isinstance(v, int):
                    data.append(f"{v}")
                elif isinstance(v, float):
                    data.append(f"{v}")
                else:
                    try:
                        # convert from pint-style values
                        data.append(f"{v.magnitude}")
                    except AttributeError:
                        data.append(str(v))  # this is the catch-all

                criteria = point[header].get("criteria")
                if criteria is not None:
                    if "pass_if" in criteria.keys():
                        data.append(f'{criteria["pass_if"]}')
                    if "min" in criteria.keys():
                        data.append(f'{criteria["min"]}')
                    if "max" in criteria.keys():
                        data.append(f'{criteria["max"]}')

        data_string = f"{self._delimiter.join(data)}\n"

        self._save_file(header_string, data_string)

    def _save_file(self, header_string: str, data_string: str):
        """
        Saves a new file if header has changed or appends to the old file.

        :param header_string: the string containing the header
        :param data_string: the string containing the data
        """
        destination_path = self._path / self._fname

        # check for the header string
        header_changed = False
        try:
            # if the header strings do not match, then set the
            # 'create_new_file' flag
            with open(destination_path, "r") as f:
                len_new_header_string = len(header_string)
                old_header_string = f.read(len_new_header_string)

                if old_header_string != header_string:
                    self._logger.info("header strings do not match")
                    header_changed = True
        except FileNotFoundError:
            # if the file is not found, then create a new file
            self._logger.info(f'file "{destination_path}" not found...')

        if header_changed:
            now = datetime.now().strftime("%Y-%m-%dT%H%M%S")

            # rename the old file using ISO8601 timestamp
            parts = self._fname.split(".")
            parts[0] = f"{parts[0]}_{now}"
            new_path = self._path / ".".join(parts)

            self._logger.info(
                f'attempting to copy "{destination_path}" ' f'to "{new_path}"...'
            )
            copy(destination_path, new_path)
            remove(destination_path)

        # write the header string
        if not destination_path.exists():
            self._logger.info(
                f'"{destination_path}" does not exist, creating' f" with new heading"
            )
            with open(destination_path, "w") as f:
                f.write(header_string)

        # write the data string
        self._logger.info(f'appending data: "{data_string.strip()}"')
        with open(destination_path, "a") as f:
            f.write(data_string)



import atexit
from datetime import datetime
import logging
from threading import Thread
import traceback
from time import sleep
from typing import Optional

from mats.test import Test
from mats.archiving import ArchiveManager

# State Machine - Using strings b/c we can relatively easily look
# for valid substrings to give more information in a concise way.
# For instance, one may be looking for the substring 'ready', but
# another function may look for the substring 'abort'.
# valid states:
#  - ready
#  - starting
#  - setting up
#  - executing tests
#  - tearing down
#  - complete / ready
#  - aborting
#  - aborted / ready
#  - exiting


class TestSequence:

    """
    The sequence or stack of ``Test`` objects to execute.

    The TestSequence will "knit" the sequence together by taking the test \
    objects and appropriately passing them through the automated testing \
    process.

    :param sequence: a list of Tests
    :param archive_manager: an instance of ``ArchiveManager`` which will \
    contain the path and data_format-specific information
    :param auto_run: an integer that determines how many times a test \
    sequence will be executed before stopping
    :param callback: function to call on each test sequence completion; \
    callback will be required to accept one parameter, which is the \
    dictionary of values collected over that test iteration
    :param setup: function to call before the test sequence
    :param teardown: function to call after the test sequence is complete, \
    even if there was a problem; common to have safety issues addressed here
    :param on_close: function to call when the functionality is complete; \
    for instance, when a GUI closes, test hardware may need to be de-allocated
    :param loglevel: the logging level
    """

    def __init__(
        self,
        sequence: list[Test],
        archive_manager: Optional[ArchiveManager] = None,
        auto_run: Optional[int] = None,
        callback: Optional[callable] = None,
        setup: Optional[callable] = None,
        teardown: Optional[callable] = None,
        on_close: Optional[callable] = None,
        loglevel=logging.INFO,
    ):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._logger.setLevel(loglevel)

        # protection just in case one or more of the instances contained
        # within the sequence were not instantiated properly, this will
        # instantiate them
        for i, test in enumerate(sequence):
            if not isinstance(test, Test):
                sequence[i] = test()

        if not TestSequence.__validate_sequence(sequence):
            raise ValueError("test monikers are not uniquely identified")

        self._sequence = sequence
        self._archive_manager = archive_manager
        self._callback = callback
        self._setup = setup
        self._teardown = teardown
        self._on_close = on_close
        self._auto_run = auto_run
        self._state = "ready" if not auto_run else "starting"

        self._test_data = {
            "datetime": str(datetime.now()),
            "pass": True,
            "failed": [],
        }

        self.current_test = None
        self._current_test_number = 0

        if self._teardown is not None:
            atexit.register(self._teardown_function)

        self._thread = Thread(target=self._run_sequence, daemon=True)
        self._thread.start()

    def __getitem__(self, test_element: (str, Test)):
        """
        Returns the ``Test`` instance within the \
        sequence named ``test_element``.

        :param test_element: a string that matches the moniker
        """
        test = None

        if isinstance(test_element, str):
            for t in self._sequence:
                if t.moniker == test_element:
                    test = t
                    break
        elif isinstance(test_element, Test):
            for t in self._sequence:
                if t == test_element:
                    test = t
                    break
        else:
            raise TypeError(
                f'"{test_element}" is of type "{type(test_element)}"; '
                f'expected "str" or "Test"'
            )

        if test is not None:
            return test

        raise KeyError(f'test "{test_element}" does not appear to be defined')

    @property
    def tests(self) -> list[Test]:
        """
        Returns instances of all tests contained within the ``TestSequence``

        :return: all tests as a list
        """
        return [test for test in self._sequence]

    @property
    def test_names(self) -> list[str]:
        """
        Returns the names of the tests contained within the ``TestSequence``

        :return: the names of the tests as a list
        """
        return [test.moniker for test in self._sequence]

    @property
    def ready(self) -> bool:
        """
        Returns True if the test sequence is ready for another go at it, \
        False if not

        :return: True or False
        """
        return "ready" in self._state

    @property
    def is_passing(self) -> bool:
        """
        Returns True if the test sequence is currently passing, else False

        :return: True or False
        """
        return self._test_data.get("pass")

    @property
    def is_aborted(self) -> bool:
        """
        Returns True if the test sequence has been aborted, else False

        :return: True or False
        """
        return "abort" in self._state

    @property
    def failed_tests(self) -> list[str]:
        """
        Return a list of the tests which failed.

        :return: list of tests that failed
        """
        return self._test_data.get("failed")

    @property
    def progress(self) -> tuple[int, int]:
        """
        Returns a tuple containing (current_test_number, total_tests) in \
        order to give an indication of the progress of the test sequence.

        :return: tuple containing (current_test_number, total_tests)
        """
        return (
            self._current_test_number,
            len([test.moniker for test in self._sequence]),
        )

    @property
    def in_progress(self) -> bool:
        """
        Returns True if the test sequence is currently in progress, else False.

        :return: True if the test sequence is currently in progress, else False
        """
        return "ready" not in self._state

    def close(self):
        """
        Allows higher level code to call the close functionality.
        """
        self._state = "exiting"
        if self._on_close is not None:
            self._on_close()

    @staticmethod
    def __validate_sequence(sequence: list[Test]):
        moniker_set = set([t.moniker for t in sequence])

        if len(moniker_set) != len(sequence):
            return False

        return True

    def abort(self):
        """
        Abort the current test sequence.

        :return: None
        """
        if "ready" not in self._state:
            self._state = "aborting"
            [test.abort() for test in self._sequence]

    def start(self):
        """
        Start a test sequence.  Will only work if a test sequence isn't \
        already in progress.

        :return: None
        """
        if self.in_progress:
            self._logger.warning(
                "cannot begin another test when test is " "currently in progress"
            )
            return

        self._state = "starting"

    def _teardown_function(self):
        self._logger.info(
            "tearing down test sequence by " "executing sequence teardown function"
        )

        try:
            self._teardown()
        except Exception as e:
            self._logger.critical(f"critical error during " f"test teardown: {e}")
            self._logger.critical(str(traceback.format_exc()))

    def _reset_sequence(self):
        """
        Initializes the test sequence data, initializes each \
        `Test` in preparation for the next single execution \
        of the sequence.
        """

    def _run_sequence(self):
        """
        Runs one instance of the test sequence.

        :return: None
        """
        while self._state != "exiting":
            # wait at the ready (unless in auto-run mode)
            while "ready" in self._state:
                if self._auto_run and "abort" not in self._state:
                    self._logger.info(
                        '"auto_run" flag is set, ' "beginning test sequence"
                    )
                    self.start()
                else:
                    sleep(0.1)

            if self._state == "exiting":
                self._sequence_teardown()
                return

            self._sequence_setup()
            self._sequence_executing_tests()
            self._sequence_teardown()

            if "abort" not in self._state:
                if self._archive_manager is not None:
                    self._archive_manager.aggregate(
                        datetime=self._test_data["datetime"],
                        is_passing=self._test_data["pass"],
                        failed=self._test_data["failed"],
                        tests=self._sequence,
                    )

            if self._auto_run:
                self._auto_run -= 1

            if "abort" in self._state:
                self._state = "aborted / ready"
            else:
                self._state = "complete / ready"

    def _sequence_setup(self):
        if "abort" in self._state:
            return

        self._state = "setting up"
        self._logger.info("-" * 80)
        self._test_data = {
            "datetime": str(datetime.now()),
            "pass": True,
            "failed": [],
        }

        self._current_test_number = 0

        for test in self._sequence:
            test.reset()

        if self._setup is not None:
            self._setup()

    def _sequence_executing_tests(self):
        if "abort" in self._state:
            return

        # begin the test sequence
        for i, test in enumerate(self._sequence):
            self._current_test_number = i

            if "abort" in self._state:
                self._logger.warning(
                    f"abort detected on test " f"{i}, exiting test sequence"
                )
                break

            self.current_test = test.moniker

            try:
                test._setup(is_passing=self.is_passing)
            except Exception as e:
                self._logger.critical(
                    f"critical error during " f'setup of "{test}": {e}'
                )
                self._logger.critical(str(traceback.format_exc()))
                self.abort()
                test.fail()

            if test.aborted:
                self.abort()
                break

            try:
                test._execute(is_passing=self.is_passing)
            except Exception as e:
                self._logger.critical(
                    f"critical error during " f'execution of "{test}": {e}'
                )
                self._logger.critical(str(traceback.format_exc()))
                self.abort()
                test.fail()

            if test.aborted:
                self.abort()
                test.fail()
                break

            try:
                test._teardown(is_passing=self.is_passing)
            except Exception as e:
                self._logger.critical(
                    f"critical error during " f'teardown of "{test}": {e}'
                )
                self._logger.critical(str(traceback.format_exc()))
                self.abort()
                test.fail()

            if test.aborted:
                self.abort()
                break

            if not test._test_is_passing:
                self._test_data["pass"] = False
                self._test_data["failed"].append(test.moniker)

        if "abort" in self._state:
            self._test_data["pass"] = None

    def _sequence_teardown(self):
        """
        Finishes up a test sequence by saving data, executing teardown \
        sequence, along with user callbacks.
        :return:
        """
        if "abort" not in self._state:
            self._state = "tearing down"

        self._logger.info("test sequence complete")
        self._logger.debug(f"test results: {self._test_data}")

        if self._teardown is not None:
            try:
                self._teardown()
            except Exception as e:
                self._logger.critical(
                    f"an exception has occurred which "
                    f"may result in an unsafe condition "
                    f"during sequence teardown: {e}"
                )

        if self._callback is not None:
            self._logger.info(
                f"executing user-supplied callback function " f'"{self._callback}"'
            )
            try:
                self._callback(self._test_data)
            except Exception as e:
                self._logger.warning(
                    f"an exception occurred during the callback sequence: {e}"
                )



"""
Automated test suite for the Automated Test Environment.  This file focuses
on testing the ``TestSequence`` class.
"""
from os import path, remove
from time import sleep

import pytest

import mats


test_counter = 0  # use this to keep track of some test execution counts


class T_normal_1(mats.Test):
    def __init__(self):
        super().__init__('test 1', pass_if=0, significant_figures=1)

    def execute(self, is_passing):
        sleep(0.2)
        return 0

    def teardown(self, is_passing):
        global test_counter
        test_counter += 1


class T_normal_2(mats.Test):
    def __init__(self):
        super().__init__('test 2',
                         min_value=-1.0,
                         max_value=1.0,
                         significant_figures=2)

    def execute(self, is_passing):
        # provide a dummy number that
        # will definitely pass the criteria
        return 0.0121


class T_aborted(mats.Test):
    def __init__(self):
        super().__init__('test aborting')

    def execute(self, is_passing):
        sleep(0.1)
        self.abort()
        return None


class T_failing(mats.Test):
    def __init__(self):
        super().__init__('test failing', pass_if=True)

    def execute(self, is_passing):
        return False


class T_setup_exception(mats.Test):
    def __init__(self):
        super().__init__('test setup exception')

    def setup(self, is_passing):
        raise ValueError

    def execute(self, is_passing):
        return False


class T_execute_exception(mats.Test):
    def __init__(self):
        super().__init__('test execute exception')

    def execute(self, is_passing):
        raise ValueError


class T_teardown_exception(mats.Test):
    def __init__(self):
        super().__init__('test teardown exception')

    def execute(self, is_passing):
        return False

    def teardown(self, is_passing):
        raise ValueError


t1, t2 = T_normal_1(), T_normal_2()
ta, tf = T_aborted(), T_failing()


@pytest.fixture
def normal_test_sequence():
    global test_counter
    yield mats.TestSequence(sequence=[t1, t2])
    test_counter = 0


@pytest.fixture
def aborted_test_sequence():
    global test_counter
    yield mats.TestSequence(sequence=[t1, ta, t2])
    test_counter = 0


@pytest.fixture
def failed_test_sequence():
    global test_counter
    yield mats.TestSequence(sequence=[t1, tf])
    test_counter = 0


@pytest.fixture
def auto_run_sequence():
    global test_counter
    yield mats.TestSequence(sequence=[t1, t2],
                            auto_run=True)
    test_counter = 0


def test_TestSequence_creation(normal_test_sequence):
    ts = normal_test_sequence

    assert ts.in_progress is False
    assert len(ts.test_names) == 2
    assert len(ts.tests) == 2


def test_TestSequence_duplicate_monikers():
    with pytest.raises(ValueError):
        mats.TestSequence(sequence=[t1, t1, t1])


def test_TestSequence_uninstantiated_Tests():
    mats.TestSequence(sequence=[
        T_normal_1, T_normal_2
    ])


def test_TestSequence_empty_raises_TE():
    ts = mats.TestSequence(sequence=[
        T_normal_1, T_normal_2
    ])
    with pytest.raises(KeyError):
        ts['test 3']


def test_TestSequence_retrieve_by_moniker(normal_test_sequence):
    ts = normal_test_sequence
    test = ts['test 1']
    assert test.moniker == 'test 1'


def test_TestSequence_retrieve_by_Test(normal_test_sequence):
    ts = normal_test_sequence
    test = ts[t1]
    assert test.moniker == 'test 1'


def test_TestSequence_retrieve_by_integer_should_error(normal_test_sequence):
    ts = normal_test_sequence
    with pytest.raises(TypeError):
        return ts[0]


def test_TestSequence_run():
    """
    Checking a normal test sequence.
    """
    ts = mats.TestSequence(sequence=[t1, t2])

    assert ts.ready
    ts.start()

    # wait a small amount of time, ensure that the test sequence has
    # begun and is in progress
    sleep(0.1)
    assert ts.in_progress is True

    while ts.in_progress is True:
        sleep(0.1)

    assert ts.in_progress is False


def test_TestSequence_close():
    """
    Checking a normal test sequence.
    """
    close_events = 0

    def on_close():
        nonlocal close_events
        close_events += 1

    ts = mats.TestSequence(sequence=[t1, t2], on_close=on_close)

    assert ts.ready
    ts.start()

    sleep(0.1)
    while ts.in_progress is True:
        sleep(0.1)

    assert close_events == 0
    ts.close()
    assert  close_events == 1


# running GUI testing appears to make pytest unstable, leaving here
# until i can decide how to handle
# def test_MatsFrame_run_default(root):
#     """
#     Same as a normal test sequence, but uses
#     a MatsFrame with default parameters.
#     """
#     ts = mats.TestSequence(sequence=[t1, t2])
#     mf = MatsFrame(parent=root, sequence=ts)
#     mf.grid()
#     sleep(0.5)
#
#     assert ts.ready
#     ts.start()
#
#     # wait a small amount of time, ensure that the test sequence has
#     # begun and is in progress
#     sleep(0.1)
#     assert ts.in_progress is True
#
#     while ts.in_progress is True:
#         sleep(0.1)
#
#     assert ts.in_progress is False
#
#     # wait for some of the frame to update so that frames may be exercised
#     sleep(1.0)
#     root.destroy()

# running GUI testing appears to make pytest unstable, leaving here
# until i can decide how to handle
# def test_MatsFrame_run_horizontal(root):
#     """
#     Same as a normal test sequence, vertical orientation
#     """
#     ts = mats.TestSequence(sequence=[t1, t2])
#     mf = MatsFrame(parent=root, sequence=ts, vertical=False)
#     mf.grid()
#     sleep(0.5)
#
#     assert ts.ready
#     ts.start()
#
#     # wait a small amount of time, ensure that the test sequence has
#     # begun and is in progress
#     sleep(0.1)
#     assert ts.in_progress is True
#
#     while ts.in_progress is True:
#         sleep(0.1)
#
#     assert ts.in_progress is False
#
#     # wait for some of the frame to update so that frames may be exercised
#     sleep(1.0)
#     root.destroy()


def test_TestSequence_run_attempted_interrupt():
    """
    Testing to ensure that an interrupted test sequence is not actually \
    interrupted.  This test should clearly show in the coverage and in the \
    logging messages.

    :return:
    """
    ts = mats.TestSequence(sequence=[t1, t2])

    assert ts.ready
    ts.start()

    # wait a small amount of time, ensure that the test sequence has
    # begun and is in progress
    sleep(0.1)
    assert ts.in_progress is True
    ts.start()

    while ts.in_progress is True:
        sleep(0.1)

    assert ts.in_progress is False
    assert ts.progress == (1, 2, )


def test_TestSequence_run_with_callback():
    """
    Testing to ensure that the callback is executed during a normal test \
    sequence.

    :param normal_test_sequence:
    :return:
    """
    count = 0

    def increment_count(data):
        nonlocal count
        count += 1

    ts = mats.TestSequence(
        sequence=[t1, t2],
        callback=increment_count)
    ts.start()

    while(ts.in_progress):
        sleep(0.1)

    assert count == 1


def test_TestSequence_run_aborted(aborted_test_sequence):
    """
    Testing to ensure that the test is aborted, the test indicates that it is \
    aborted and that the test sequence indicates that it is aborted.

    :param aborted_test_sequence:
    :return:
    """
    ts = aborted_test_sequence

    assert ts.ready
    ts.start()

    while ts.in_progress is True:
        sleep(0.1)

    aborted_test = ts['test aborting']
    assert aborted_test.aborted is True

    assert ts.is_aborted is True

# #testing GUI elements appears to make automated testing unstable
# def test_MatsFrame_run_aborted(root):
#     """
#     Same as a normal test sequence, but uses
#     a MatsFrame with default parameters.
#     """
#     ts = mats.TestSequence(sequence=[t1, ta, t2])
#     mf = mats.MatsFrame(parent=root, sequence=ts)
#     mf.grid()
#     sleep(0.5)
#
#     assert ts.ready
#     ts.start()
#
#     # wait a small amount of time, ensure that the test sequence has
#     # begun and is in progress
#     sleep(0.1)
#     assert ts.in_progress is True
#
#     while ts.in_progress is True:
#         sleep(0.1)
#
#     aborted_test = ts['test aborting']
#     assert aborted_test.aborted is True
#
#     assert ts.is_aborted is True
#     sleep(1.0)
#     root.destroy()


def test_TestSequence_run_failed(failed_test_sequence):
    """
    Ensures that a test sequence that fails will cause the test sequence \
    to fail but not indicate aborted.

    :param failed_test_sequence:
    :return:
    """
    ts = failed_test_sequence

    assert ts.ready
    ts.start()

    while ts.in_progress is True:
        sleep(0.1)

    assert ts.is_passing is False
    assert ts.is_aborted is False
    assert len(ts.failed_tests) == 1


def test_TestSequence_setup_exception():
    """
    Tests the case in which there is some exception that occurs during \
    the setup phase of the test.  THe test should execute the teardown
    sequence.

    :return: None
    """
    counter = 0

    def on_test_exception():
        nonlocal counter
        counter += 1

    ts = mats.TestSequence(sequence=[T_setup_exception()],
                           teardown=on_test_exception)
    ts.start()

    while ts.in_progress is True:
        sleep(0.1)

    assert counter == 1


def test_TestSequence_execute_exception():
    """
    Tests the case in which there is some exception that occurs during \
    the execution phase of the test.  THe test should execute the teardown \
    sequence.

    :return: None
    """
    counter = 0

    def on_test_exception():
        nonlocal counter
        counter += 1

    ts = mats.TestSequence(sequence=[T_execute_exception()],
                           teardown=on_test_exception)
    ts.start()

    while ts.in_progress is True:
        sleep(0.1)

    assert counter == 1


def test_TestSequence_setup():
    """Ensures that the "setup" function is called when specified."""
    def setup_function():
        with open('test_data.txt', 'w') as f:
            f.write('data')

    ts = mats.TestSequence(sequence=[t1, t2],
                           setup=setup_function)
    ts.start()
    while ts.in_progress is True:
        sleep(0.1)

    assert path.exists('test_data.txt')
    remove('test_data.txt')



"""
Automated test suite for the Automated Test Environment.  This file focuses
on testing the ``Test`` class.
"""
import pytest
import mats


@pytest.fixture
def blank_Test():
    class T(mats.Test):
        def __init__(self):
            super().__init__('test')

        def execute(self, is_passing):
            return None

    yield T()


@pytest.fixture
def pass_if_Test():
    class T(mats.Test):
        def __init__(self):
            super().__init__('test', pass_if=True)

        def execute(self, is_passing):
            return None

    yield T()


@pytest.fixture
def min_Test():
    class T(mats.Test):
        def __init__(self):
            super().__init__('test', min_value=1.0)

        def execute(self, is_passing):
            return None

    yield T()


@pytest.fixture
def max_Test():
    class T(mats.Test):
        def __init__(self):
            super().__init__('test', max_value=2.0)

        def execute(self, is_passing):
            return None

    yield T()


@pytest.fixture
def bracketed_Test():
    class T(mats.Test):
        def __init__(self):
            super().__init__('test', min_value=1.0, max_value=2.0)

        def execute(self, is_passing):
            return None

    yield T()


@pytest.fixture
def aborted_in_setup_Test():
    class T(mats.Test):
        def __init__(self):
            super().__init__('test')

        def setup(self, is_passing):
            self.abort()

        def execute(self, is_passing):
            return None

    yield T()


@pytest.fixture
def aborted_in_execute_Test():
    class T(mats.Test):
        def __init__(self):
            super().__init__('test')

        def execute(self, is_passing):
            self.abort()
            return None

    yield T()


@pytest.fixture
def aborted_in_teardown_Test():
    class T(mats.Test):
        def __init__(self):
            super().__init__('test')

        def execute(self, is_passing):
            return None

        def teardown(self, is_passing):
            self.abort()

    yield T()


def test_ate_version():
    assert mats.__version__


def test_Test_execute_not_implemented():
    t = mats.Test('test')
    with pytest.raises(NotImplementedError):
        t._execute(is_passing=True)


def test_Test_creation(blank_Test):
    t = blank_Test

    assert t.moniker == 'test'
    assert t.status == 'waiting'


def test_Test_class_creation_with_pass_if(pass_if_Test):
    t = pass_if_Test

    assert t.moniker == 'test'
    assert t.criteria['pass_if'] is True


def test_Test_class_creation_min_only(min_Test):
    t = min_Test

    assert t.moniker == 'test'
    assert t.criteria['min'] == 1


def test_Test_class_creation_max_only(max_Test):
    t = max_Test

    assert t.moniker == 'test'
    assert t.criteria['max'] == 2


def test_Test_creation_with_min_and_max(bracketed_Test):
    t = bracketed_Test

    assert t.moniker == 'test'
    assert t.criteria['min'] == 1
    assert t.criteria['max'] == 2


def test_Test_blank_setup_execute_teardown(blank_Test):
    t = blank_Test

    # check values before _setup()
    assert t.is_passing is None
    assert t.status == 'waiting'

    t._setup(is_passing=True)

    # check values after _setup()
    assert t.is_passing is True
    assert t.status == 'running'

    t._execute(is_passing=True)

    # check values after _exeecute()
    assert t.is_passing is True
    assert t.status == 'running'

    t._teardown(is_passing=True)

    # check values after _teardown()
    assert t.is_passing is True
    assert t.status == 'complete'


def test_Test_pass_if_should_pass(pass_if_Test):
    t = pass_if_Test

    def execute(is_passing):
        return True

    # overwrite the `execute` method with the method above
    t.execute = execute

    t._setup(is_passing=True)
    t._execute(is_passing=True)

    assert t.is_passing is True


def test_Test_pass_if_should_fail(pass_if_Test):
    t = pass_if_Test

    def execute(is_passing):
        return False

    # overwrite the `execute` method with the method above
    t.execute = execute

    t._setup(is_passing=True)
    t._execute(is_passing=True)

    assert t.is_passing is False


def test_Test_min_should_pass(bracketed_Test):
    t = bracketed_Test

    def execute(is_passing):
        return 1.000001

    t.execute = execute

    t._setup(is_passing=True)
    t._execute(is_passing=True)

    assert t.is_passing is True


def test_Test_min_should_fail(bracketed_Test):
    t = bracketed_Test

    def execute(is_passing):
        return 0.9991

    t.execute = execute

    t._setup(is_passing=True)
    t._execute(is_passing=True)

    assert t.is_passing is False


def test_Test_max_should_pass(bracketed_Test):
    t = bracketed_Test

    def execute(is_passing):
        return 1.99999

    t.execute = execute

    t._setup(is_passing=True)
    t._execute(is_passing=True)

    assert t.is_passing is True


def test_Test_max_should_fail(bracketed_Test):
    t = bracketed_Test

    def execute(is_passing):
        return 2.001

    t.execute = execute

    t._setup(is_passing=True)
    t._execute(is_passing=True)

    assert t.is_passing is False


def test_Test_aborted_in_setup(aborted_in_setup_Test):
    t = aborted_in_setup_Test
    t._setup(is_passing=True)
    assert t.aborted is True

    t._execute(is_passing=False)
    assert t.aborted is True

    t._teardown(is_passing=False)
    assert t.aborted is True


def test_Test_aborted_in_execute(aborted_in_execute_Test):
    t = aborted_in_execute_Test

    t._setup(is_passing=True)
    assert t.aborted is False

    t._execute(is_passing=True)
    assert t.aborted is True

    t._teardown(is_passing=True)
    assert t.aborted is True


def test_Test_aborted_in_teardown(aborted_in_teardown_Test):
    t = aborted_in_teardown_Test
    t._setup(is_passing=True)
    assert t.aborted is False

    t._execute(is_passing=True)
    assert t.aborted is False

    t._teardown(is_passing=True)
    assert t.aborted is True


def test_Test_adding_extra_data(bracketed_Test):
    t = bracketed_Test

    def execute(is_passing):
        t.save_dict({'a': 1.0})
        return 1.99999

    t.execute = execute

    t._setup(is_passing=True)
    t._execute(is_passing=True)
    t._teardown(is_passing=True)

    assert t.is_passing is True
    assert t.saved_data['a'] == 1.0


def test_Test_significant_figures():
    class T(mats.Test):
        def __init__(self):
            super().__init__('test', significant_figures=4)

        def execute(self, is_passing):
            return 12345678

    test = T()
    test._execute(True)
    assert test.value == 12350000


def test_Test_significant_figures_string():
    """ if passed a string, then the significant figures should not alter it """
    class T(mats.Test):
        def __init__(self):
            super().__init__('test', significant_figures=4)

        def execute(self, is_passing):
            return '12345678'

    test = T()
    test._execute(True)
    assert test.value == '12345678'



"""
Manufacturing Automated Test System (MATS)

MATS takes care of all of the low-level tedium that every
manufacturing automated test must implement, such as:

 - test runner
 - test definition
 - auto saving of data
 - hardware setup / teardown
 - automatic GUI integration
"""

import coloredlogs
import logging

from mats.archiving import ArchiveManager

from mats.test import Test
from mats.test_sequence import TestSequence
from mats.tkwidgets import MatsFrame
from mats.version import __version__

__all__ = ["Test", "TestSequence", "ArchiveManager", "MatsFrame", "__version__"]

coloredlogs.install(level="DEBUG")

logger = logging.getLogger(__name__)



import logging
from numbers import Number
from typing import Optional, Union

from sigfig import round


class Test:

    """
    The most basic unit of an executing test sequence.

    Within a test, we may have a setup(), execute(), and \
    teardown() method.  Only the `execute()` method is required \
    to be overridden.

    :param moniker: a shortcut name for this particular test
    :param min_value: the minimum value that is to be considered a pass, \
    if defined
    :param max_value: the maximum value that is to be considered a pass, \
    if defined
    :param pass_if: the value that must be present in order to pass, if defined
    :param significant_figures: the number of significant figures appropriate to the measurement
    :param loglevel: the logging level to apply such as `logging.INFO`
    """

    def __init__(
        self,
        moniker: str,
        min_value: Optional[Number] = None,
        max_value: Optional[Number] = None,
        pass_if: Optional[Union[str, bool, int]] = None,
        significant_figures: int = 4,
        loglevel=logging.INFO,
    ):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._logger.setLevel(loglevel)

        criteria = {}
        if pass_if is not None:
            criteria["pass_if"] = pass_if
        if min_value is not None:
            criteria["min"] = min_value
        if max_value is not None:
            criteria["max"] = max_value

        self.moniker = moniker
        self._criteria = criteria if criteria else None
        self._significant_figures = significant_figures

        self._test_is_passing = None
        self.value = None
        self.aborted = False
        self.status = "waiting"

        self.saved_data = {}

    @property
    def is_passing(self):
        """
        Returns `True` if test is currently passing, else `False`

        :return: `True` if passing, else `False`
        """
        return self._test_is_passing

    @property
    def criteria(self):
        """
        Returns the test criteria as a `dict`

        :return: test criteria as a `dict`
        """
        return self._criteria

    def abort(self):
        """
        Causes current test status to abort

        :return: None
        """
        self.aborted = True

    def _setup(self, is_passing: bool):
        """
        Pre-execution method used for logging and housekeeping.

        :param is_passing: True if the test sequence is passing up to this \
        point, else False
        :return:
        """
        self._logger.info(f'setting up "{self.moniker}"')

        self._test_is_passing = True
        self.value = None
        self.status = "running" if not self.aborted else "aborted"

        self.aborted = False
        self.setup(is_passing=is_passing)
        self.status = "running" if not self.aborted else "aborted"

    def _execute(self, is_passing: bool):
        """
        Pre-execution method used for logging and housekeeping.

        :param is_passing: True if the test sequence is passing up to this \
        point, else False
        :return:
        """
        self.status = "running" if not self.aborted else "aborted"
        if self.aborted:
            self._logger.warning("aborted, not executing")
            return

        self._logger.info(f'executing test "{self.moniker}"')

        # execute the test and perform appropriate rounding
        value = self.execute(is_passing=is_passing)
        if isinstance(value, Number):
            try:
                value = round(value, self._significant_figures)
            except ValueError:
                self._logger.debug(f'could not apply significant digits to "{value}"')
        self.value = value

        if self._criteria is not None:
            if self._criteria.get("pass_if") is not None:
                if self.value != self._criteria["pass_if"]:
                    self._logger.warning(
                        f'"{self.value}" != pass_if requirement '
                        f'"{self._criteria["pass_if"]}", failing'
                    )
                    self.fail()
                else:
                    self._logger.info(
                        f'"{self.value}" == pass_if requirement '
                        f'"{self._criteria["pass_if"]}"'
                    )

            if self._criteria.get("min") is not None:
                if self.value < self._criteria["min"]:
                    self._logger.warning(
                        f'"{self.value}" is below the minimum '
                        f'"{self._criteria["min"]}", failing'
                    )
                    self.fail()
                else:
                    self._logger.info(
                        f'"{self.value}" is above the minimum '
                        f'"{self._criteria["min"]}"'
                    )

            if self._criteria.get("max") is not None:
                if self.value > self._criteria["max"]:
                    self._logger.warning(
                        f'"{self.value}" is above the maximum '
                        f'"{self._criteria["max"]}"'
                    )
                    self.fail()
                else:
                    self._logger.info(
                        f'"{self.value}" is below the maximum '
                        f'"{self._criteria["max"]}"'
                    )

        self.status = "running" if not self.aborted else "aborted"

        return self.value

    def _teardown(self, is_passing: bool):
        """
        Pre-execution method used for logging and housekeeping.

        :param is_passing: True if the test sequence is passing up to \
        this point, else False
        :return:
        """
        if self.aborted:
            self.status = "aborted"
            self._logger.warning("aborted, not executing")
            return

        self._logger.info(f'tearing down "{self.moniker}"')

        self.teardown(is_passing)
        self.status = "complete"

    def reset(self):
        """
        Reset the test status
        :return: None
        """
        self.status = "waiting"

    def save_dict(self, data: dict):
        """
        Allows a test to save additional data other than that returned \
        by ``execute()``

        :param data: key: value pairs of the data to be stored
        :return: None
        """
        self.saved_data = data.copy()

    def fail(self):
        """
        When called, will cause the test to fail.

        :return: None
        """
        self._test_is_passing = False

    def setup(self, is_passing: bool):
        """
        Abstract method intended to be overridden by subclass

        :param is_passing: True if the test sequence is passing up to this \
        point, else False
        :return: None
        """
        pass

    def execute(self, is_passing: bool):
        """
        Abstract method intended to be overridden by subclass

        :param is_passing: True if the test sequence is passing up to this \
        point, else False
        :return: value to be appended to the sequence dictionary
        """
        raise NotImplementedError

    def teardown(self, is_passing: bool):
        """
        Abstract method intended to be overridden by subclass

        :param is_passing: True if the test sequence is passing up to this \
        point, else False
        :return: None
        """
        pass



import logging
from tkinter import Frame, Button, Label

from mats.test import Test
from mats.test_sequence import TestSequence

_light_green = "#66ff66"
_light_red = "#ff6666"
_light_yellow = "#ffff99"
_relief = "sunken"
_label_padding = 5


class MatsFrame(Frame):

    """
    The frame that interacts with the test sequence to display the \
    test results as the test is executing.

    :param parent: the tk parent frame
    :param sequence: the instance of `TestSequence` to monitor
    :param vertical: if `True`, will stack tests vertically; \
        otherwise, horizontally; default is vertical, `True`
    :param start_btn: if `True`, will populate a start button; \
        otherwise, will not; default is `True`
    :param abort_btn: if `True`, will populate an abort button; \
        otherwise, will not; default is `True`
    :param wrap: the number above which will start the next row or column
    :param loglevel: the logging level, for instance 'logging.INFO'
    """

    def __init__(
        self,
        parent,
        sequence: TestSequence,
        vertical: bool = True,
        start_btn: bool = True,
        abort_btn: bool = True,
        wrap: int = 6,
        loglevel=logging.INFO,
    ):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._logger.setLevel(loglevel)

        self._parent = parent
        super().__init__(self._parent)

        self._sequence = sequence

        arrow = "\u2b9e" if not vertical else "\u2b9f"

        row = 0
        col = 0

        if start_btn or abort_btn:
            btn_frame = Frame(self)
            btn_frame.grid(row=0, column=0, sticky="news")
            btn_frame.columnconfigure(0, weight=1)
            r = 0
            if start_btn:
                Button(btn_frame, text="Start", command=sequence.start).grid(
                    row=r, column=0, sticky="news"
                )
                r += 1

            if abort_btn:
                Button(btn_frame, text="ABORT", command=sequence.abort, fg="red").grid(
                    row=r, column=0, sticky="news"
                )

        status_frame = Frame(self)
        if vertical:
            status_frame.grid(row=1, column=0)
        else:
            Label(self, text=arrow, justify="center", anchor="center").grid(
                row=0, column=1, sticky="news"
            )
            status_frame.grid(row=0, column=2)

        col += 1 if not vertical else 0
        row += 1 if vertical else 0

        self._test_status_frames = []
        for test in self._sequence.tests:
            self._test_status_frames.append(
                _TestLabel(
                    status_frame, test, loglevel=self._logger.getEffectiveLevel()
                )
            )

        row, col = 0, 0
        max_row = row
        max_col = col
        for tl in self._test_status_frames:
            if vertical:
                tl.grid(row=row, column=col, sticky="news")

                row += 1
                row %= wrap
                if row == 0:
                    col += 1
            else:
                tl.grid(row=row, column=col, sticky="news")

                col += 1
                col %= wrap
                if col == 0:
                    row += 1

            max_row = max(row, max_row)
            max_col = max(col, max_col)

        col += 1 if not vertical else 0
        row += 1 if vertical else 0

        self._complete_label = Label(self, text="-", anchor="center", justify="center")
        if vertical:
            self._complete_label.grid(row=2, column=0, sticky="news")
        else:
            self._complete_label.grid(row=0, column=3, sticky="news")
        self._complete_label.config(
            relief=_relief, padx=_label_padding, pady=_label_padding
        )

        self._update()

    def _update(self):
        if self._sequence.in_progress:
            self._complete_label.config(text="in progress", background=_light_yellow)
        elif self._sequence.is_aborted:
            self._complete_label.config(text="aborted", background=_light_red)
        elif self._sequence.is_passing:
            self._complete_label.config(text="pass", background=_light_green)
        else:
            self._complete_label.config(text="fail", background=_light_red)

        self.after(100, self._update)


class _TestLabel(Label):

    """
    A single instance of a test label frame.
    """

    def __init__(self, parent, test: Test, loglevel=logging.INFO):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._logger.setLevel(loglevel)

        self._parent = parent
        super().__init__(self._parent)

        self._test = test

        try:
            criteria = self._test.criteria
        except AttributeError:
            criteria = None
        criteria_list = []
        criteria_string = ""
        if criteria is not None:
            for condition, value in criteria.items():
                if (
                    isinstance(value, bool)
                    or isinstance(value, int)
                    or isinstance(value, str)
                ):
                    cs = f"{condition}={value}"
                else:
                    cs = f"{condition}={value}"
                criteria_list.append(cs)
            criteria_string = ",".join(criteria_list)

        self._label_text = f"{self._test.moniker}"
        if criteria_string:
            self._label_text += f"\n{criteria_string}"

        self.config(
            text=self._label_text,
            relief=_relief,
            padx=_label_padding,
            pady=_label_padding,
        )

        self._label_bg_color = self.cget("background")

        self._update()

    def _update(self):
        """
        Updates the test label display based on the status of the Test

        :return: None
        """
        color = self._label_bg_color

        if self._test.status == "waiting":
            color = self._label_bg_color
        elif self._test.status == "running":
            color = _light_yellow
        elif self._test.status == "aborted":
            color = _light_red
        elif not self._test.is_passing:
            color = _light_red
        elif self._test.is_passing:
            color = _light_green

        value = self._test.value
        if isinstance(value, float):
            value_str = f"{value}"
        else:
            try:
                value_str = f"{value.magnitude}"
            except AttributeError:
                value_str = f"{value}"

        max_length = 12
        if value_str and len(value_str) <= max_length:
            label_text = f"{self._label_text}\n({value_str.strip()})"
        else:
            self._logger.info(
                "the value string length is greater than "
                f"{max_length} and, thus, will not be shown on the GUI"
            )
            label_text = self._label_text

        self.config(background=color, text=label_text)

        self.after(100, self._update)



"""
Automated test suite for the Automated Test Environment.

This file focuses on testing the ``ArchiveManager`` class.
"""
from pathlib import Path
from os import remove

import pint
import pytest

import mats

unit = pint.UnitRegistry()


data_point_1 = {'t1': {'value': 10},
                't2': {'value': 10.0},
                't3': {'value': 'string 10'},
                't5': {'value': True}}
data_point_2 = {'t1': {'value': 10},
                't2': {'value': 10.0},
                't3': {'value': 'string 10'},
                't4': {'value': 1 * unit.rpm},
                't5': {'value': True}}
data_point_3 = {'t1': {'value': 10},
                't2': {'value': 10.0, 'criteria': {'min': 9.0, 'max': 11.0}},
                't3': {'value': 'string 10'},
                't4': {'value': 1 * unit.rpm},
                't5': {'value': True, 'criteria': {'pass_if': True}}}


@pytest.fixture
def am0():
    path = Path('.')

    yield mats.ArchiveManager(path=path, data_format=0)

    data_paths = [f for f in Path('.').iterdir() if 'data' in str(f)]
    for p in data_paths:
        remove(p)


@pytest.fixture
def am1():
    path = Path('.')

    yield mats.ArchiveManager(path=path, data_format=1)

    data_paths = [f for f in Path('.').iterdir() if 'data' in str(f)]
    for p in data_paths:
        remove(p)


@pytest.fixture
def am2():
    path = Path('.')

    yield mats.ArchiveManager(path=path, data_format=2)

    data_paths = [f for f in Path('.').iterdir() if 'data' in str(f)]
    for p in data_paths:
        remove(p)


def test_am_creation(am1, am2):
    assert am1._delimiter == '\t'
    assert am2._delimiter == '\t'


def test_am0_save(am0):
    length = 5

    for _ in range(length):
        am0.save(data_point_1)

    assert Path('data.txt').exists

    with open(Path('data.txt'), 'r') as f:
        assert len(f.readlines()) == (length + 2)


def test_am1_save(am1):
    length = 5

    for _ in range(length):
        am1.save(data_point_1)

    assert Path('data.txt').exists

    with open(Path('data.txt'), 'r') as f:
        assert len(f.readlines()) == (length + 1)


def test_am2_save(am2):
    length = 5

    for _ in range(length):
        with pytest.raises(ValueError):
            am2.save(data_point_1)


def test_am0_save_with_criteria(am0):
    am0.save(data_point_3)

    data = data_point_3.copy()
    data['t5']['value'] = False
    am0.save(data_point_3)

    with Path('data.txt').open('r') as f:
        assert len(f.readlines()) == 6


def test_am1_save_with_criteria(am1):
    am1.save(data_point_3)

    data = data_point_3.copy()
    data['t5']['value'] = False
    am1.save(data_point_3)

    with Path('data.txt').open('r') as f:
        assert len(f.readlines()) == 3


def test_am0_save_new_header(am0):
    length = 5

    for _ in range(length):
        am0.save(data_point_1)

    for _ in range(length):
        am0.save(data_point_2)

    data_paths = [f for f in Path('.').iterdir() if 'data' in str(f)]
    assert len(data_paths) == 2

    for p in data_paths:
        with open(p, 'r') as f:
            assert len(f.readlines()) == (length + 2)


def test_am1_save_new_header(am1):
    length = 5

    for _ in range(length):
        am1.save(data_point_1)

    for _ in range(length):
        am1.save(data_point_2)

    data_paths = [f for f in Path('.').iterdir() if 'data' in str(f)]
    assert len(data_paths) == 2

    for p in data_paths:
        with open(p, 'r') as f:
            assert len(f.readlines()) == (length + 1)
